@@ -98,8 +98,17 @@ on_object_added (GDBusObjectManager *manager,
                  GDBusObject *object,
                  gpointer user_data)
 {
-        if (is_realm_with_kerberos_and_membership (object))
+        GList *interfaces, *l;
+
+        interfaces = g_dbus_object_get_interfaces (object);
+        for (l = interfaces; l != NULL; l = g_list_next (l))
+                on_interface_added (manager, object, l->data);
+        g_list_free_full (interfaces, g_object_unref);
+
+        if (is_realm_with_kerberos_and_membership (object)) {
+                g_debug ("Saw realm: %s", g_dbus_object_get_object_path (object));
                 g_signal_emit (user_data, signals[REALM_ADDED], 0, object);
+        }
 }
 
 static void
@@ -159,36 +168,87 @@ on_realm_diagnostics (GDBusConnection *connection,
         }
 }
 
-static gboolean
-number_at_least (const gchar *number,
-                 guint minimum)
-{
-        gchar *end;
+typedef struct {
+        GCancellable *cancellable;
+        UmRealmManager *manager;
+} NewClosure;
 
-        if (strtol (number, &end, 10) < (long)minimum)
-                return FALSE;
-        if (!end || *end != '\0')
-                return FALSE;
-        return TRUE;
+static void
+new_closure_free (gpointer data)
+{
+        NewClosure *closure = data;
+        g_clear_object (&closure->cancellable);
+        g_clear_object (&closure->manager);
+        g_slice_free (NewClosure, closure);
 }
 
-static gboolean
-version_compare (const char *version,
-                 guint req_major,
-                 guint req_minor)
+static void
+on_provider_new (GObject *source,
+                 GAsyncResult *result,
+                 gpointer user_data)
 {
-        gboolean match = FALSE;
-        gchar **parts;
+        GSimpleAsyncResult *async = G_SIMPLE_ASYNC_RESULT (user_data);
+        NewClosure *closure = g_simple_async_result_get_op_res_gpointer (async);
+        GError *error = NULL;
+        UmRealmProvider *provider;
 
-        parts = g_strsplit (version, ".", 2);
+        provider = um_realm_provider_proxy_new_finish (result, &error);
+        closure->manager->provider = provider;
 
-        if (parts[0] && parts[1]) {
-                match = number_at_least (parts[0], req_major) &&
-                        number_at_least (parts[1], req_minor);
+        if (error == NULL) {
+                g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (closure->manager->provider), -1);
+                g_debug ("Created realm manager");
+        } else {
+                g_simple_async_result_take_error (async, error);
+        }
+        g_simple_async_result_complete (async);
+
+        g_object_unref (async);
+}
+
+static void
+on_manager_new (GObject *source,
+                GAsyncResult *result,
+                gpointer user_data)
+{
+        GSimpleAsyncResult *async = G_SIMPLE_ASYNC_RESULT (user_data);
+        NewClosure *closure = g_simple_async_result_get_op_res_gpointer (async);
+        GDBusConnection *connection;
+        GError *error = NULL;
+        GObject *object;
+        guint sig;
+
+        object = g_async_initable_new_finish (G_ASYNC_INITABLE (source), result, &error);
+        if (error == NULL) {
+                closure->manager = UM_REALM_MANAGER (object);
+                connection = g_dbus_object_manager_client_get_connection (G_DBUS_OBJECT_MANAGER_CLIENT (object));
+
+                g_debug ("Connected to realmd");
+
+                sig = g_dbus_connection_signal_subscribe (connection,
+                                                          "org.freedesktop.realmd",
+                                                          "org.freedesktop.realmd.Service",
+                                                          "Diagnostics",
+                                                          NULL,
+                                                          NULL,
+                                                          G_DBUS_SIGNAL_FLAGS_NONE,
+                                                          on_realm_diagnostics,
+                                                          NULL,
+                                                          NULL);
+                closure->manager->diagnostics_sig = sig;
+
+                um_realm_provider_proxy_new (connection,
+                                             G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
+                                             "org.freedesktop.realmd",
+                                             "/org/freedesktop/realmd",
+                                             closure->cancellable,
+                                             on_provider_new, g_object_ref (async));
+        } else {
+                g_simple_async_result_take_error (async, error);
+                g_simple_async_result_complete (async);
         }
 
-        g_strfreev (parts);
-        return match;
+        g_object_unref (async);
 }
 
 void
@@ -196,74 +256,45 @@ um_realm_manager_new (GCancellable *cancellable,
                       GAsyncReadyCallback callback,
                       gpointer user_data)
 {
+        GSimpleAsyncResult *async;
+        NewClosure *closure;
+
+        g_debug ("Connecting to realmd...");
+
+        async = g_simple_async_result_new (NULL, callback, user_data,
+                                           um_realm_manager_new);
+        closure = g_slice_new (NewClosure);
+        closure->cancellable = cancellable ? g_object_ref (cancellable) : NULL;
+        g_simple_async_result_set_op_res_gpointer (async, closure, new_closure_free);
+
         g_async_initable_new_async (UM_TYPE_REALM_MANAGER, G_PRIORITY_DEFAULT,
-                                    cancellable, callback, user_data,
+                                    cancellable, on_manager_new, g_object_ref (async),
                                     "flags", G_DBUS_OBJECT_MANAGER_CLIENT_FLAGS_NONE,
                                     "name", "org.freedesktop.realmd",
                                     "bus-type", G_BUS_TYPE_SYSTEM,
                                     "object-path", "/org/freedesktop/realmd",
                                     "get-proxy-type-func", um_realm_object_manager_client_get_proxy_type,
                                     NULL);
+
+        g_object_unref (async);
 }
 
 UmRealmManager *
 um_realm_manager_new_finish (GAsyncResult *result,
                              GError **error)
 {
-        UmRealmManager *self;
-        GDBusConnection *connection;
-        GObject *source_object;
-        const gchar *version;
-        GObject *ret;
-        guint sig;
+        GSimpleAsyncResult *async;
+        NewClosure *closure;
 
-        source_object = g_async_result_get_source_object (result);
-        ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), result, error);
-        g_object_unref (source_object);
+        g_return_val_if_fail (g_simple_async_result_is_valid (result, NULL,
+                                                              um_realm_manager_new), NULL);
 
-        if (ret == NULL)
+        async = G_SIMPLE_ASYNC_RESULT (result);
+        if (g_simple_async_result_propagate_error (async, error))
                 return NULL;
 
-        self = UM_REALM_MANAGER (ret);
-        connection = g_dbus_object_manager_client_get_connection (G_DBUS_OBJECT_MANAGER_CLIENT (self));
-
-        /*
-         * TODO: Remove this version checking. This is temporary code, so
-         * just use sync here. Shortly we'll be stabilizing the realmd
-         * interfaces.
-         */
-
-        self->provider = um_realm_provider_proxy_new_sync (connection, G_DBUS_PROXY_FLAGS_NONE,
-                                                           "org.freedesktop.realmd",
-                                                           "/org/freedesktop/realmd",
-                                                           NULL, error);
-        if (self->provider == NULL) {
-                g_object_unref (self);
-                return NULL;
-        }
-
-        version = um_realm_provider_get_version (self->provider);
-        if (version == NULL || !version_compare (version, 0, 7)) {
-                /* No need to bother translators with this temporary message */
-                g_set_error (error, UM_REALM_ERROR, UM_REALM_ERROR_GENERIC,
-                             "realmd version should be at least 0.7");
-                g_object_unref (self);
-                return NULL;
-        }
-
-        sig = g_dbus_connection_signal_subscribe (connection,
-                                                  "org.freedesktop.realmd",
-                                                  "org.freedesktop.realmd.Service",
-                                                  "Diagnostics",
-                                                  NULL,
-                                                  NULL,
-                                                  G_DBUS_SIGNAL_FLAGS_NONE,
-                                                  on_realm_diagnostics,
-                                                  NULL,
-                                                  NULL);
-        self->diagnostics_sig = sig;
-
-        return self;
+        closure = g_simple_async_result_get_op_res_gpointer (async);
+        return g_object_ref (closure->manager);
 }
 
 typedef struct {
@@ -291,6 +322,7 @@ on_provider_discover (GObject *source,
         DiscoverClosure *discover = g_simple_async_result_get_op_res_gpointer (async);
         GDBusObject *object;
         GError *error = NULL;
+        gboolean no_membership = FALSE;
         gchar **realms;
         gint relevance;
         gint i;
@@ -300,17 +332,30 @@ on_provider_discover (GObject *source,
         if (error == NULL) {
                 for (i = 0; realms[i]; i++) {
                         object = g_dbus_object_manager_get_object (discover->manager, realms[i]);
-                        if (object == NULL)
+                        if (object == NULL) {
                                 g_warning ("Realm is not in object manager: %s", realms[i]);
-                        else
-                                discover->realms = g_list_prepend (discover->realms, object);
+                        } else {
+                                if (is_realm_with_kerberos_and_membership (object)) {
+                                        g_debug ("Discovered realm: %s", realms[i]);
+                                        discover->realms = g_list_prepend (discover->realms, object);
+                                } else {
+                                        g_debug ("Realm does not support kerberos membership: %s", realms[i]);
+                                        no_membership = TRUE;
+                                        g_object_unref (object);
+                                }
+                        }
                 }
+                g_strfreev (realms);
 
+                if (!discover->realms && no_membership) {
+                        g_simple_async_result_set_error (async, UM_REALM_ERROR, UM_REALM_ERROR_GENERIC,
+                                                         _("Cannot automatically join this type of domain"));
+                }
         } else {
                 g_simple_async_result_take_error (async, error);
-                g_simple_async_result_complete (async);
         }
 
+        g_simple_async_result_complete (async);
         g_object_unref (async);
 }
 
@@ -329,6 +374,8 @@ um_realm_manager_discover (UmRealmManager *self,
         g_return_if_fail (input != NULL);
         g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
 
+        g_debug ("Discovering realms for: %s", input);
+
         res = g_simple_async_result_new (G_OBJECT (self), callback, user_data,
                                          um_realm_manager_discover);
         discover = g_slice_new0 (DiscoverClosure);
@@ -475,6 +522,9 @@ on_realm_join_complete (GObject *source,
                         gpointer user_data)
 {
 	GSimpleAsyncResult *async = G_SIMPLE_ASYNC_RESULT (user_data);
+
+	g_debug ("Completed Join() method call");
+
 	g_simple_async_result_set_op_res_gpointer (async, g_object_ref (result), g_object_unref);
 	g_simple_async_result_complete_in_idle (async);
 	g_object_unref (async);
@@ -494,6 +544,7 @@ realm_join_as_owner (UmRealmObject *realm,
         GSimpleAsyncResult *async;
         GVariant *contents;
         GVariant *options;
+        GVariant *option;
         GVariant *creds;
         const gchar *type;
 
@@ -502,6 +553,7 @@ realm_join_as_owner (UmRealmObject *realm,
 
         type = find_supported_credentials (membership, owner);
         if (type == NULL) {
+                g_debug ("Couldn't find supported credential type for owner: %s", owner);
                 g_object_unref (membership);
                 return FALSE;
         }
@@ -510,12 +562,14 @@ realm_join_as_owner (UmRealmObject *realm,
                                            realm_join_as_owner);
 
         if (g_str_equal (type, "ccache")) {
+                g_debug ("Using a kerberos credential cache to join the realm");
                 contents = g_variant_new_from_data (G_VARIANT_TYPE ("ay"),
                                                     g_bytes_get_data (credentials, NULL),
                                                     g_bytes_get_size (credentials),
                                                     TRUE, (GDestroyNotify)g_bytes_unref, credentials);
 
         } else if (g_str_equal (type, "password")) {
+                g_debug ("Using a user/password to join the realm");
                 contents = g_variant_new ("(ss)", login, password);
 
         } else {
@@ -523,7 +577,10 @@ realm_join_as_owner (UmRealmObject *realm,
         }
 
         creds = g_variant_new ("(ssv)", type, owner, contents);
-        options = g_variant_new_array (G_VARIANT_TYPE ("{sv}"), NULL, 0);
+        option = g_variant_new ("{sv}", "manage-system", g_variant_new_boolean (FALSE));
+        options = g_variant_new_array (G_VARIANT_TYPE ("{sv}"), &option, 1);
+
+        g_debug ("Calling the Join() method with %s credentials", owner);
 
         um_realm_kerberos_membership_call_join (membership, creds, options,
                                                 cancellable, on_realm_join_complete,
@@ -600,6 +657,7 @@ um_realm_join_finish (UmRealmObject *realm,
 
         dbus_error = g_dbus_error_get_remote_error (call_error);
         if (dbus_error == NULL) {
+                g_debug ("Join() failed because of %s", call_error->message);
                 g_propagate_error (error, call_error);
                 return FALSE;
         }
@@ -607,10 +665,17 @@ um_realm_join_finish (UmRealmObject *realm,
         g_dbus_error_strip_remote_error (call_error);
 
         if (g_str_equal (dbus_error, "org.freedesktop.realmd.Error.AuthenticationFailed")) {
+                g_debug ("Join() failed because of invalid/insufficient credentials");
                 g_set_error (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_LOGIN,
                              "%s", call_error->message);
                 g_error_free (call_error);
+        } else if (g_str_equal (dbus_error, "org.freedesktop.realmd.Error.BadHostname")) {
+                g_debug ("Join() failed because of invalid/conflicting host name");
+                g_set_error (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_HOSTNAME,
+                             "%s", call_error->message);
+                g_error_free (call_error);
         } else {
+                g_debug ("Join() failed because of %s", call_error->message);
                 g_propagate_error (error, call_error);
         }
 
@@ -654,10 +719,16 @@ login_perform_kinit (krb5_context k5,
 
         name = g_strdup_printf ("%s@%s", login, realm);
         code = krb5_parse_name (k5, name, &principal);
-        g_free (name);
 
-        if (code != 0)
+        if (code != 0) {
+                g_debug ("Couldn't parse principal name: %s: %s",
+                         name, krb5_get_error_message (k5, code));
+                g_free (name);
                 return code;
+        }
+
+        g_debug ("Using principal name to kinit: %s", name);
+        g_free (name);
 
         if (filename == NULL)
                 code = krb5_cc_default (k5, &ccache);
@@ -666,6 +737,9 @@ login_perform_kinit (krb5_context k5,
 
         if (code != 0) {
                 krb5_free_principal (k5, principal);
+                g_debug ("Couldn't open credential cache: %s: %s",
+                         filename ? filename : "<default>",
+                         krb5_get_error_message (k5, code));
                 return code;
         }
 
@@ -683,18 +757,23 @@ login_perform_kinit (krb5_context k5,
         krb5_cc_close (k5, ccache);
         krb5_free_principal (k5, principal);
 
-        if (code == 0)
+        if (code == 0) {
+                g_debug ("kinit succeeded");
                 krb5_free_cred_contents (k5, &creds);
+        } else {
+                g_debug ("kinit failed: %s", krb5_get_error_message (k5, code));
+        }
 
         return code;
 }
 
 static void
-kinit_thread_func (GSimpleAsyncResult *async,
-                   GObject *object,
+kinit_thread_func (GTask *task,
+                   gpointer source_object,
+                   gpointer task_data,
                    GCancellable *cancellable)
 {
-        LoginClosure *login = g_simple_async_result_get_op_res_gpointer (async);
+        LoginClosure *login = task_data;
         krb5_context k5 = NULL;
         krb5_error_code code;
         GError *error = NULL;
@@ -727,35 +806,46 @@ kinit_thread_func (GSimpleAsyncResult *async,
                         g_file_get_contents (filename, &contents, &length, &error);
                         if (error == NULL) {
                                 login->credentials = g_bytes_new_take (contents, length);
+                                g_debug ("Read in credential cache: %s", filename);
                         } else {
-                                g_warning ("Couldn't read credential cache: %s", error->message);
+                                g_warning ("Couldn't read credential cache: %s: %s",
+                                           filename, error->message);
                                 g_error_free (error);
                         }
                 }
+                g_task_return_boolean (task, TRUE);
                 break;
 
         case KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN:
-        case KRB5KDC_ERR_CLIENT_REVOKED:
-        case KRB5KDC_ERR_KEY_EXP:
         case KRB5KDC_ERR_POLICY:
-        case KRB5KDC_ERR_ETYPE_NOSUPP:
-                g_simple_async_result_set_error (async, UM_REALM_ERROR, UM_REALM_ERROR_BAD_LOGIN,
-                                                 _("Cannot log in as %s at the %s domain"),
-                                                 login->user, login->domain);
+                g_task_return_new_error (task, UM_REALM_ERROR, UM_REALM_ERROR_BAD_LOGIN,
+                                         _("Cannot log in as %s at the %s domain"),
+                                         login->user, login->domain);
                 break;
         case KRB5KDC_ERR_PREAUTH_FAILED:
-                g_simple_async_result_set_error (async, UM_REALM_ERROR, UM_REALM_ERROR_BAD_PASSWORD,
-                                                 _("Invalid password, please try again"));
+        case KRB5KRB_AP_ERR_BAD_INTEGRITY:
+                g_task_return_new_error (task, UM_REALM_ERROR, UM_REALM_ERROR_BAD_PASSWORD,
+                                         _("Invalid password, please try again"));
+                break;
+        case KRB5_PREAUTH_FAILED:
+        case KRB5KDC_ERR_KEY_EXP:
+        case KRB5KDC_ERR_CLIENT_REVOKED:
+        case KRB5KDC_ERR_ETYPE_NOSUPP:
+        case KRB5_PROG_ETYPE_NOSUPP:
+                g_task_return_new_error (task, UM_REALM_ERROR, UM_REALM_ERROR_CANNOT_AUTH,
+                                         _("Cannot log in as %s at the %s domain"),
+                                         login->user, login->domain);
                 break;
         default:
-                g_simple_async_result_set_error (async, UM_REALM_ERROR, UM_REALM_ERROR_GENERIC,
-                                                 _("Couldn't connect to the %s domain: %s"),
-                                                 login->domain, krb5_get_error_message (k5, code));
+                g_task_return_new_error (task, UM_REALM_ERROR, UM_REALM_ERROR_GENERIC,
+                                         _("Couldn't connect to the %s domain: %s"),
+                                         login->domain, krb5_get_error_message (k5, code));
                 break;
         }
 
         if (filename) {
                 g_unlink (filename);
+                g_debug ("Deleted credential cache: %s", filename);
                 g_free (filename);
         }
 
@@ -771,7 +861,7 @@ um_realm_login (UmRealmObject *realm,
                 GAsyncReadyCallback callback,
                 gpointer user_data)
 {
-        GSimpleAsyncResult *async;
+        GTask *task;
         LoginClosure *login;
         UmRealmKerberos *kerberos;
 
@@ -783,40 +873,40 @@ um_realm_login (UmRealmObject *realm,
         kerberos = um_realm_object_get_kerberos (realm);
         g_return_if_fail (kerberos != NULL);
 
-        async = g_simple_async_result_new (NULL, callback, user_data,
-                                           um_realm_login);
+        task = g_task_new (realm, cancellable, callback, user_data);
         login = g_slice_new0 (LoginClosure);
         login->domain = g_strdup (um_realm_kerberos_get_domain_name (kerberos));
         login->realm = g_strdup (um_realm_kerberos_get_realm_name (kerberos));
         login->user = g_strdup (user);
         login->password = g_strdup (password);
-        g_simple_async_result_set_op_res_gpointer (async, login, login_closure_free);
+        g_task_set_task_data (task, login, login_closure_free);
 
-        g_simple_async_result_set_handle_cancellation (async, TRUE);
-        g_simple_async_result_run_in_thread (async, kinit_thread_func,
-                                             G_PRIORITY_DEFAULT, cancellable);
+        g_task_set_check_cancellable (task, TRUE);
+        g_task_set_return_on_cancel (task, TRUE);
 
-        g_object_unref (async);
+        g_task_run_in_thread (task, kinit_thread_func);
+
+        g_object_unref (task);
         g_object_unref (kerberos);
 }
 
 gboolean
-um_realm_login_finish (GAsyncResult *result,
+um_realm_login_finish (UmRealmObject *realm,
+                       GAsyncResult *result,
                        GBytes **credentials,
                        GError **error)
 {
-        GSimpleAsyncResult *async;
+        GTask *task;
         LoginClosure *login;
 
-        g_return_val_if_fail (g_simple_async_result_is_valid (result, NULL,
-                              um_realm_login), FALSE);
+        g_return_val_if_fail (g_task_is_valid (result, realm), FALSE);
         g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
-        async = G_SIMPLE_ASYNC_RESULT (result);
-        if (g_simple_async_result_propagate_error (async, error))
+        task = G_TASK (result);
+        if (!g_task_propagate_boolean (task, error))
                 return FALSE;
 
-        login = g_simple_async_result_get_op_res_gpointer (async);
+        login = g_task_get_task_data (task);
         if (credentials) {
                 if (login->credentials)
                         *credentials = g_bytes_ref (login->credentials);
